using System.Collections.Generic;
using System.IO;

using VirtualLab.Programming;
using VirtualLabAPI.Core.Common;
using VirtualLabAPI.Core.FieldRepresentations;
using VirtualLabAPI.Core.Numerics;
using VirtualLabAPI.Core.OpticalSystems;
using VirtualLabAPI.Core.StandardTransmissionDesign;

using VirtualLabAPI.Core.Detectors;

// Version 1.1c for VirtualLab 2024.1 or higher.

namespace OwnCode {
	public class VLModule {
		//the path where all the data is located
		string pathofIFTAInputData = @"D:\Data\IFTA Example";
		//file name of the IFTA document which should be loaded from hard disc
		string filenameIFTA = "Iterative Fourier Transform Algorithm Optimization.dp";
		//file name of the target pattern that shall be load into the IFTA
		string filenameTargetField = "LT_Logo.ca2";
		//define filename for storage of merit function values
		string filenameMeritFunctionValues = "Result.txt";

		//define number of steps for each iteration
		int numberSteps_PhaseSynthesis = 25;
		int numberSteps_SNROptimization = 50;
		int numberSteps_SoftQuantization = 100;
		int numberSteps_SRNforQuantization = 5000;

		bool limitScaleFactor = true;
		double scaleFactorLimit = 1;

		public void Run() {
			//load IFTA from hard disc
			DesignAlgorithmHandler design = DesignAlgorithmHandler.Load(Path.Combine(pathofIFTAInputData, filenameIFTA));
			//error handling
			if (design == null) {
				Globals.DataDisplay.LogError("IFTA could not be loaded!");
				return;
			}

			//load signal field
			ComplexAmplitude caSignalField = ComplexAmplitude.Load(Path.Combine(pathofIFTAInputData, filenameTargetField));

			//error handling
			if (caSignalField == null) {
				Globals.DataDisplay.LogError("Signal could not be loaded!");
				return;
			}

			//set signal field to IFTA document
			design.ConstraintSpecification.SetSignalFieldAndSignalRegion(caSignalField.Field, null);

			//define number of steps
			design.NumItSPO = numberSteps_PhaseSynthesis;
			design.NumItSNRPhase = numberSteps_SNROptimization;
			design.NumItSoftQuant = numberSteps_SoftQuantization;
			design.NumItSNRQuant = numberSteps_SRNforQuantization;

			//set parameter for limit scale factor
			design.ConstraintSpecification.LimitScaleFactor = limitScaleFactor;
			design.ConstraintSpecification.ScaleFactorLimitGoalEff = scaleFactorLimit;

			//read sampling parameters from design document
			SamplingParameters sPara = new SamplingParameters();
			sPara = new SamplingParameters(design.ConstraintSpecification.SamplingPoints,
										   design.ConstraintSpecification.SamplingDistance);

			if (design.Transmission != null) {
				//show initial transmission
				ComplexAmplitude initialTransmission = new ComplexAmplitude(design.Transmission);
				initialTransmission.SamplingDistance = sPara.SamplingDistance;
				Globals.DataDisplay.ShowDocument(initialTransmission, "Initial Transmission");
			}

			//perform design
			IterationDataOutput iterationDataOutput = new IterationDataOutput();
			design.DoAllSelectedSteps(iterationDataOutput, null);

			//show optimized transmission
			ComplexAmplitude optimizedTransmission = new ComplexAmplitude(design.Transmission);
			optimizedTransmission.SamplingDistance = sPara.SamplingDistance;
			Globals.DataDisplay.ShowDocument(optimizedTransmission, "Optimized Transmission");

			SamplingParameters sParaOutPutField = new SamplingParameters();
			sParaOutPutField = new SamplingParameters(design.ConstraintSpecification.SamplingPoints,
													  design.ConstraintSpecification.SamplingDistanceOut);
			//design.ConstraintSpecification.
			ComplexField cfOutputField = design.DoForwardProp();
			//show optimized transmission
			ComplexAmplitude resultField = new ComplexAmplitude(cfOutputField);
			resultField.SamplingDistance = sParaOutPutField.SamplingDistance;
			Globals.DataDisplay.ShowDocument(resultField, "Optimized Output Field");

			DiffractiveOpticsMeritFunctionsCalculation resultCalculation = GetMeritFunctionCalculation(design, cfOutputField, sParaOutPutField.SamplingDistance);
			resultCalculation.CalculateMeritFunctions([StandardMeritFunction.ParaxialConversionEfficiency, StandardMeritFunction.UniformityError]);
			List<PhysicalValue> results = resultCalculation.GetPhysicalValues();

			Globals.DataDisplay.LogDetectorResult("IFTA Module", results);

			Globals.DataDisplay.ShowDocument(design, "Modified Design (performed!)");

			//define path for storing merit functions
			string filenameMeritFunction = Path.Combine(pathofIFTAInputData, filenameMeritFunctionValues);
			//check whether file already exist
			if (File.Exists(filenameMeritFunction)) {
				File.Delete(filenameMeritFunction);
			}

			VL_Files.WriteLineToLogFile(filenameMeritFunction, results[0].ToString("F")); // "F" ensures that also the description of the physical value is written.
			VL_Files.WriteLineToLogFile(filenameMeritFunction, results[1].ToString("F"));
		}

		/// <summary>
		/// Gets the merit function calculation object.
		/// </summary>
		/// <param name="design">The design for which the merit functions are to be calculated.</param>
		/// <param name="outField">The output field.</param>
		/// <param name="samplingDistanceOut">The sampling distance of the output field.</param>
		/// <returns>The merit function calculation object. You still need to call the matching Calculate method.</returns>
		public DiffractiveOpticsMeritFunctionsCalculation GetMeritFunctionCalculation(DesignAlgorithmHandler design, 
																					  ComplexField outField, 
																					  VectorD samplingDistanceOut) {
			double sumNormIn = design.ConstraintSpecification.GetSumNormInputField();

			return new(outField, sumNormIn, samplingDistanceOut,
					   design.ConstraintSpecification.SignalField,
					   design.ConstraintSpecification.GetRealSignalRegion(),
					   design.ConstraintSpecification.AllowScaleFreedom,
					   design.ConstraintSpecification.AllowPhaseFreedom);
		}		
	}
}
